@@ -1,6 +1,6 @@
 // Copyright 2008-present Contributors to the OpenImageIO project.
 // SPDX-License-Identifier: BSD-3-Clause
-// https://github.com/OpenImageIO/oiio/blob/master/LICENSE.md
+// https://github.com/OpenImageIO/oiio
 
 // clang-format off
 
@@ -14,12 +14,26 @@
 #include <OpenImageIO/simd.h>
 #include <OpenImageIO/ustring.h>
 #include <OpenImageIO/varyingref.h>
+#include <OpenImageIO/vecparam.h>
 
 
 // Define symbols that let client applications determine if newly added
 // features are supported.
 #define OIIO_TEXTURESYSTEM_SUPPORTS_CLOSE 1
 
+// Is the getattributetype() method present? (Added in 2.5)
+#define OIIO_TEXTURESYSTEM_SUPPORTS_GETATTRIBUTETYPE 1
+
+#define OIIO_TEXTURESYSTEM_SUPPORTS_STOCHASTIC 1
+
+#ifndef INCLUDED_IMATHVEC_H
+// Placeholder declaration for Imath::V3f if no Imath headers have been
+// included.
+namespace Imath {
+template <class T> class Vec3;
+using V3f = Vec3<float>;
+}
+#endif
 
 
 OIIO_NAMESPACE_BEGIN
@@ -93,7 +107,9 @@ enum class MipMode {
     NoMIP,      ///< Just use highest-res image, no MIP mapping
     OneLevel,   ///< Use just one mipmap level
     Trilinear,  ///< Use two MIPmap levels (trilinear)
-    Aniso       ///< Use two MIPmap levels w/ anisotropic
+    Aniso,      ///< Use two MIPmap levels w/ anisotropic
+    StochasticTrilinear, ///< DEPRECATED Stochastic trilinear
+    StochasticAniso, ///< DEPRECATED Stochastic anisotropic
 };
 
 /// Interp mode determines how we sample within a mipmap level
@@ -102,7 +118,7 @@ enum class InterpMode {
     Closest,      ///< Force closest texel
     Bilinear,     ///< Force bilinear lookup within a mip level
     Bicubic,      ///< Force cubic lookup within a mip level
-    SmartBicubic  ///< Bicubic when maxifying, else bilinear
+    SmartBicubic  ///< Bicubic when magnifying, else bilinear
 };
 
 
@@ -115,8 +131,8 @@ enum class InterpMode {
 /// The SIMD width for batched texturing operations. This is fixed within
 /// any release of OpenImageIO, but may change from release to release and
 /// also may be overridden at build time. A typical batch size is 16.
-static constexpr int BatchWidth = OIIO_TEXTURE_SIMD_BATCH_WIDTH;
-static constexpr int BatchAlign = BatchWidth * sizeof(float);
+OIIO_INLINE_CONSTEXPR int BatchWidth = OIIO_TEXTURE_SIMD_BATCH_WIDTH;
+OIIO_INLINE_CONSTEXPR int BatchAlign = BatchWidth * sizeof(float);
 
 /// A type alias for a SIMD vector of floats with the batch width.
 typedef simd::VecType<float, OIIO_TEXTURE_SIMD_BATCH_WIDTH>::type FloatWide;
@@ -135,15 +151,15 @@ typedef uint64_t RunMask;
 // The defined constant `RunMaskOn` contains the value with all bits
 // `0..BatchWidth-1` set to 1.
 #if OIIO_TEXTURE_SIMD_BATCH_WIDTH == 4
-static constexpr RunMask RunMaskOn = 0xf;
+OIIO_INLINE_CONSTEXPR RunMask RunMaskOn = 0xf;
 #elif OIIO_TEXTURE_SIMD_BATCH_WIDTH == 8
-static constexpr RunMask RunMaskOn = 0xff;
+OIIO_INLINE_CONSTEXPR RunMask RunMaskOn = 0xff;
 #elif OIIO_TEXTURE_SIMD_BATCH_WIDTH == 16
-static constexpr RunMask RunMaskOn = 0xffff;
+OIIO_INLINE_CONSTEXPR RunMask RunMaskOn = 0xffff;
 #elif OIIO_TEXTURE_SIMD_BATCH_WIDTH == 32
-static constexpr RunMask RunMaskOn = 0xffffffff;
+OIIO_INLINE_CONSTEXPR RunMask RunMaskOn = 0xffffffff;
 #elif OIIO_TEXTURE_SIMD_BATCH_WIDTH == 64
-static constexpr RunMask RunMaskOn = 0xffffffffffffffffULL;
+OIIO_INLINE_CONSTEXPR RunMask RunMaskOn = 0xffffffffffffffffULL;
 #else
 #    error "Not a valid OIIO_TEXTURE_SIMD_BATCH_WIDTH choice"
 #endif
@@ -190,7 +206,9 @@ public:
         MipModeNoMIP,      ///< Just use highest-res image, no MIP mapping
         MipModeOneLevel,   ///< Use just one mipmap level
         MipModeTrilinear,  ///< Use two MIPmap levels (trilinear)
-        MipModeAniso       ///< Use two MIPmap levels w/ anisotropic
+        MipModeAniso,      ///< Use two MIPmap levels w/ anisotropic
+        MipModeStochasticTrilinear, ///< DEPRECATED Stochastic trilinear
+        MipModeStochasticAniso, ///< DEPRECATED Stochastic anisotropic
     };
 
     /// Interp mode determines how we sample within a mipmap level
@@ -199,7 +217,7 @@ public:
         InterpClosest,      ///< Force closest texel
         InterpBilinear,     ///< Force bilinear lookup within a mip level
         InterpBicubic,      ///< Force cubic lookup within a mip level
-        InterpSmartBicubic  ///< Bicubic when maxifying, else bilinear
+        InterpSmartBicubic  ///< Bicubic when magnifying, else bilinear
     };
 
 
@@ -212,10 +230,8 @@ public:
         anisotropic(32), conservative_filter(true),
         sblur(0.0f), tblur(0.0f), swidth(1.0f), twidth(1.0f),
         fill(0.0f), missingcolor(nullptr),
-        // dresultds(nullptr), dresultdt(nullptr),
-        time(0.0f), bias(0.0f), samples(1),
-        rwrap(WrapDefault), rblur(0.0f), rwidth(1.0f), // dresultdr(nullptr),
-        // actualchannels(0),
+        time(0.0f), rnd(-1.0f), samples(1),
+        rwrap(WrapDefault), rblur(0.0f), rwidth(1.0f),
         envlayout(0)
     { }
 
@@ -237,7 +253,10 @@ public:
     float fill;                 ///< Fill value for missing channels
     const float* missingcolor;  ///< Color for missing texture
     float time;                 ///< Time (for time-dependent texture lookups)
-    float bias;                 ///< Bias for shadows
+    union {
+        float bias;                 ///< Bias for shadows (DEPRECATED)
+        float rnd;                  ///< Stratified sample value
+    };
     int samples;                ///< Number of samples for shadows
 
     // For 3D volume texture lookups only:
@@ -281,7 +300,10 @@ class OIIO_API TextureOptBatch {
 public:
     /// Create a TextureOptBatch with all fields initialized to reasonable
     /// defaults.
-    TextureOptBatch () {}   // use inline initializers
+    TextureOptBatch () {
+        for (int i = 0; i < Tex::BatchWidth; ++i)
+            rnd[i] = -1.0f;
+    }
 
     // Options that may be different for each point we're texturing
     alignas(Tex::BatchAlign) float sblur[Tex::BatchWidth];    ///< Blur amount
@@ -291,6 +313,9 @@ public:
     alignas(Tex::BatchAlign) float twidth[Tex::BatchWidth];
     alignas(Tex::BatchAlign) float rwidth[Tex::BatchWidth];
     // Note: rblur,rwidth only used for volumetric lookups
+#if OIIO_VERSION_GREATER_EQUAL(2,4,0)
+    alignas(Tex::BatchAlign) float rnd[Tex::BatchWidth];
+#endif
 
     // Options that must be the same for all points we're texturing at once
     int firstchannel = 0;                 ///< First channel of the lookup
@@ -316,13 +341,13 @@ private:
 
 
 
-/// DEPRECATED(1.8)
-/// Encapsulate all the options needed for texture lookups.  Making
-/// these options all separate parameters to the texture API routines is
-/// very ugly and also a big pain whenever we think of new options to
-/// add.  So instead we collect all those little options into one
-/// structure that can just be passed by reference to the texture API
-/// routines.
+// DEPRECATED(1.8)
+// Encapsulate all the options needed for texture lookups.  Making
+// these options all separate parameters to the texture API routines is
+// very ugly and also a big pain whenever we think of new options to
+// add.  So instead we collect all those little options into one
+// structure that can just be passed by reference to the texture API
+// routines.
 class OIIO_API TextureOptions {
 public:
     /// Wrap mode describes what happens when texture coordinates describe
@@ -354,15 +379,17 @@ public:
         InterpClosest,      ///< Force closest texel
         InterpBilinear,     ///< Force bilinear lookup within a mip level
         InterpBicubic,      ///< Force cubic lookup within a mip level
-        InterpSmartBicubic  ///< Bicubic when maxifying, else bilinear
+        InterpSmartBicubic  ///< Bicubic when magnifying, else bilinear
     };
 
     /// Create a TextureOptions with all fields initialized to reasonable
     /// defaults.
+    OIIO_DEPRECATED("no longer used since OIIO 1.8")
     TextureOptions();
 
     /// Convert a TextureOpt for one point into a TextureOptions with
     /// uniform values.
+    OIIO_DEPRECATED("no longer used since OIIO 1.8")
     TextureOptions(const TextureOpt& opt);
 
     // Options that must be the same for all points we're texturing at once
@@ -518,6 +545,9 @@ public:
     ///             How many times to retry a read failure.
     /// - `int deduplicate` :
     ///             If nonzero, detect duplicate textures (default=1).
+    /// - `int max_open_files_strict` :
+    ///             If nonzero, work harder to make sure that we have
+    ///             smaller possible overages to the max open files limit.
     /// - `string substitute_image` :
     ///             If supplied, an image to substatute for all texture
     ///             references.
@@ -571,6 +601,11 @@ public:
     /// - `int flip_t` :
     ///             If nonzero, `t` coordinates will be flipped `1-t` for
     ///             all texture lookups. The default is 0.
+    /// - `int stochastic` :
+    ///             Bit field determining how to use stochastic sampling for
+    ///             MipModeStochasticAniso and/or MipModeStochasticTrilinear.
+    ///             Bit 1 = sample MIP level, bit 2 = sample anisotropy
+    ///             (default=0).
     ///
     /// - `string options`
     ///             This catch-all is simply a comma-separated list of
@@ -686,6 +721,12 @@ public:
     /// as a `std::string`.
     virtual bool getattribute(string_view name, std::string& val) const = 0;
 
+    /// If the named attribute is known, return its data type. If no such
+    /// attribute exists, return `TypeUnknown`.
+    ///
+    /// This was added in version 2.5.
+    virtual TypeDesc getattributetype (string_view name) const = 0;
+
     /// @}
 
     /// @{
@@ -749,18 +790,29 @@ public:
     /// any internals.
     class TextureHandle;
 
-    /// Retrieve an opaque handle for fast texture lookups.  The opaque
-    /// pointer `thread_info` is thread-specific information returned by
-    /// `get_perthread_info()`.  Return nullptr if something has gone
-    /// horribly wrong.
-    virtual TextureHandle * get_texture_handle (ustring filename,
+    /// Retrieve an opaque handle for fast texture lookups.  The filename is
+    /// presumed to be UTF-8 encoded. The opaque pointer `thread_info` is
+    /// thread-specific information returned by `get_perthread_info()`.
+    /// Return nullptr if something has gone horribly wrong.
+    virtual TextureHandle* get_texture_handle (ustring filename,
                                             Perthread *thread_info=nullptr) = 0;
+    /// Get a TextureHandle using a UTF-16 encoded wstring filename.
+    TextureHandle* get_texture_handle (const std::wstring& filename,
+                                       Perthread *thread_info=nullptr) {
+        return get_texture_handle (ustring(Strutil::utf16_to_utf8(filename)),
+                                   thread_info);
+    }
 
     /// Return true if the texture handle (previously returned by
     /// `get_image_handle()`) is a valid texture that can be subsequently
     /// read.
     virtual bool good(TextureHandle* texture_handle) = 0;
 
+    /// Given a handle, return the UTF-8 encoded filename for that texture.
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual ustring filename_from_handle(TextureHandle* handle) = 0;
+
     /// @}
 
     /// @{
@@ -785,7 +837,7 @@ public:
     /// receive an antialiased texture lookup.
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encoded ustring.
     /// @param  options
     ///     Fields within `options` that are honored for 2D texture lookups
     ///     include the following:
@@ -873,7 +925,7 @@ public:
     /// the volume file itself.
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encoded ustring.
     /// @param  options
     ///     Fields within `options` that are honored for 3D texture lookups
     ///     include the following:
@@ -940,8 +992,8 @@ public:
     ///             plugin.
     ///
     virtual bool texture3d (ustring filename, TextureOpt &options,
-                            const Imath::V3f &P, const Imath::V3f &dPdx,
-                            const Imath::V3f &dPdy, const Imath::V3f &dPdz,
+                            V3fParam P, V3fParam dPdx,
+                            V3fParam dPdy, V3fParam dPdz,
                             int nchannels, float *result,
                             float *dresultds=nullptr, float *dresultdt=nullptr,
                             float *dresultdr=nullptr) = 0;
@@ -950,8 +1002,8 @@ public:
     /// a texture handle and per-thread info.
     virtual bool texture3d (TextureHandle *texture_handle,
                             Perthread *thread_info, TextureOpt &options,
-                            const Imath::V3f &P, const Imath::V3f &dPdx,
-                            const Imath::V3f &dPdy, const Imath::V3f &dPdz,
+                            V3fParam P, V3fParam dPdx,
+                            V3fParam dPdy, V3fParam dPdz,
                             int nchannels, float *result,
                             float *dresultds=nullptr, float *dresultdt=nullptr,
                             float *dresultdr=nullptr) = 0;
@@ -962,16 +1014,16 @@ public:
     // Return true if the file is found and could be opened by an
     // available ImageIO plugin, otherwise return false.
     virtual bool shadow (ustring filename, TextureOpt &options,
-                         const Imath::V3f &P, const Imath::V3f &dPdx,
-                         const Imath::V3f &dPdy, float *result,
+                         V3fParam P, V3fParam dPdx,
+                         V3fParam dPdy, float *result,
                          float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
 
     // Slightly faster version of texture3d() lookup if the app already
     // has a texture handle and per-thread info.
     virtual bool shadow (TextureHandle *texture_handle, Perthread *thread_info,
                          TextureOpt &options,
-                         const Imath::V3f &P, const Imath::V3f &dPdx,
-                         const Imath::V3f &dPdy, float *result,
+                         V3fParam P, V3fParam dPdx,
+                         V3fParam dPdy, float *result,
                          float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
 
     /// Perform a filtered directional environment map lookup in the
@@ -980,7 +1032,7 @@ public:
     /// stored in `result[]`.
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  options
     ///     Fields within `options` that are honored for environment lookups
     ///     include the following:
@@ -1031,16 +1083,16 @@ public:
     ///             found or could not be opened by any available ImageIO
     ///             plugin.
     virtual bool environment (ustring filename, TextureOpt &options,
-                              const Imath::V3f &R, const Imath::V3f &dRdx,
-                              const Imath::V3f &dRdy, int nchannels, float *result,
+                              V3fParam R, V3fParam dRdx,
+                              V3fParam dRdy, int nchannels, float *result,
                               float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
 
     /// Slightly faster version of environment() if the app already has a
     /// texture handle and per-thread info.
     virtual bool environment (TextureHandle *texture_handle,
                               Perthread *thread_info, TextureOpt &options,
-                              const Imath::V3f &R, const Imath::V3f &dRdx,
-                              const Imath::V3f &dRdy, int nchannels, float *result,
+                              V3fParam R, V3fParam dRdx,
+                              V3fParam dRdy, int nchannels, float *result,
                               float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
 
     /// @}
@@ -1062,7 +1114,7 @@ public:
     /// required).
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  options
     ///             A TextureOptBatch containing texture lookup options.
     ///             This is conceptually the same as a TextureOpt, but the
@@ -1113,8 +1165,10 @@ public:
                           float *dresultds=nullptr,
                           float *dresultdt=nullptr) = 0;
 
+#ifndef OIIO_DOXYGEN
     // Old multi-point API call.
     // DEPRECATED (1.8)
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool texture (ustring filename, TextureOptions &options,
                           Runflag *runflags, int beginactive, int endactive,
                           VaryingRef<float> s, VaryingRef<float> t,
@@ -1122,6 +1176,7 @@ public:
                           VaryingRef<float> dsdy, VaryingRef<float> dtdy,
                           int nchannels, float *result,
                           float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool texture (TextureHandle *texture_handle,
                           Perthread *thread_info, TextureOptions &options,
                           Runflag *runflags, int beginactive, int endactive,
@@ -1130,6 +1185,7 @@ public:
                           VaryingRef<float> dsdy, VaryingRef<float> dtdy,
                           int nchannels, float *result,
                           float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
+#endif
 
     /// Perform filtered 3D volumetric texture lookups on a batch of
     /// positions from the same texture, all at once. The "point-like"
@@ -1146,7 +1202,7 @@ public:
     /// batch values for channel 1, etc.
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  options
     ///             A TextureOptBatch containing texture lookup options.
     ///             This is conceptually the same as a TextureOpt, but the
@@ -1197,8 +1253,10 @@ public:
                             float *dresultds=nullptr, float *dresultdt=nullptr,
                             float *dresultdr=nullptr) = 0;
 
+#ifndef OIIO_DOXYGEN
     // Retrieve a 3D texture lookup at many points at once.
     // DEPRECATED(1.8)
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool texture3d (ustring filename, TextureOptions &options,
                             Runflag *runflags, int beginactive, int endactive,
                             VaryingRef<Imath::V3f> P,
@@ -1208,6 +1266,7 @@ public:
                             int nchannels, float *result,
                             float *dresultds=nullptr, float *dresultdt=nullptr,
                             float *dresultdr=nullptr) = 0;
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool texture3d (TextureHandle *texture_handle,
                             Perthread *thread_info, TextureOptions &options,
                             Runflag *runflags, int beginactive, int endactive,
@@ -1218,6 +1277,7 @@ public:
                             int nchannels, float *result,
                             float *dresultds=nullptr, float *dresultdt=nullptr,
                             float *dresultdr=nullptr) = 0;
+#endif
 
     /// Perform filtered directional environment map lookups on a batch of
     /// directions from the same texture, all at once. The "point-like"
@@ -1234,7 +1294,7 @@ public:
     /// batch values for channel 1, etc.
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  options
     ///             A TextureOptBatch containing texture lookup options.
     ///             This is conceptually the same as a TextureOpt, but the
@@ -1280,9 +1340,11 @@ public:
                               int nchannels, float *result,
                               float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
 
+#ifndef OIIO_DOXYGEN
     // Retrieve an environment map lookup for direction R, for many
     // points at once.
     // DEPRECATED(1.8)
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool environment (ustring filename, TextureOptions &options,
                               Runflag *runflags, int beginactive, int endactive,
                               VaryingRef<Imath::V3f> R,
@@ -1290,6 +1352,7 @@ public:
                               VaryingRef<Imath::V3f> dRdy,
                               int nchannels, float *result,
                               float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool environment (TextureHandle *texture_handle,
                               Perthread *thread_info, TextureOptions &options,
                               Runflag *runflags, int beginactive, int endactive,
@@ -1298,6 +1361,7 @@ public:
                               VaryingRef<Imath::V3f> dRdy,
                               int nchannels, float *result,
                               float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
+#endif
 
     // Batched shadow lookups
     virtual bool shadow (ustring filename,
@@ -1309,8 +1373,10 @@ public:
                          const float *P, const float *dPdx, const float *dPdy,
                          float *result, float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
 
+#ifndef OIIO_DOXYGEN
     // Retrieve a shadow lookup for position P at many points at once.
     // DEPRECATED(1.8)
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool shadow (ustring filename, TextureOptions &options,
                          Runflag *runflags, int beginactive, int endactive,
                          VaryingRef<Imath::V3f> P,
@@ -1318,6 +1384,7 @@ public:
                          VaryingRef<Imath::V3f> dPdy,
                          float *result,
                          float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
+    OIIO_DEPRECATED("no longer support this multi-point call (1.8)")
     virtual bool shadow (TextureHandle *texture_handle, Perthread *thread_info,
                          TextureOptions &options,
                          Runflag *runflags, int beginactive, int endactive,
@@ -1326,6 +1393,7 @@ public:
                          VaryingRef<Imath::V3f> dPdy,
                          float *result,
                          float *dresultds=nullptr, float *dresultdt=nullptr) = 0;
+#endif
 
     /// @}
 
@@ -1334,8 +1402,8 @@ public:
     /// @name   Texture metadata and raw texels
     ///
 
-    /// Given possibly-relative 'filename', resolve it using the search
-    /// path rules and return the full resolved filename.
+    /// Given possibly-relative 'filename' (UTF-8 encoded), resolve it using
+    /// the search path rules and return the full resolved filename.
     virtual std::string resolve_filename (const std::string &filename) const=0;
 
     /// Get information or metadata about the named texture and store it in
@@ -1499,7 +1567,7 @@ public:
     ///
     ///
     /// @param  filename
-    ///             The name of the texture.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  subimage
     ///             The subimage to query. (The metadata retrieved is for
     ///             the highest-resolution MIP level of that subimage.)
@@ -1535,7 +1603,7 @@ public:
     /// subimage by default, or as set by `subimage`).
     ///
     /// @param  filename
-    ///             The name of the image.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  subimage
     ///             The subimage to query. (The spec retrieved is for the
     ///             highest-resolution MIP level of that subimage.)
@@ -1569,7 +1637,7 @@ public:
     /// underlying ImageCache.
     ///
     /// @param  filename
-    ///             The name of the image.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  subimage
     ///             The subimage to query.  (The spec retrieved is for the
     ///             highest-resolution MIP level of that subimage.)
@@ -1597,7 +1665,7 @@ public:
     /// `options.fill` value.
     ///
     /// @param  filename
-    ///             The name of the image.
+    ///             The name of the texture, as a UTF-8 encode ustring.
     /// @param  options
     ///             A TextureOpt describing access options, including wrap
     ///             modes, fill value, and subimage, that will be used when
@@ -1643,27 +1711,84 @@ public:
 
     /// @}
 
+    /// @{
+    /// @name Methods for UDIM patterns
+    ///
+
+    /// Is the UTF-8 encoded filename a UDIM pattern?
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual bool is_udim(ustring filename) = 0;
+
+    /// Does the handle refer to a file that's a UDIM pattern?
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual bool is_udim(TextureHandle* udimfile) = 0;
+
+    /// For a UDIM filename pattern (UTF-8 encoded) and texture coordinates,
+    /// return the TextureHandle pointer for the concrete tile file it refers
+    /// to, or nullptr if there is no corresponding tile (udim sets are
+    /// allowed to be sparse).
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual TextureHandle* resolve_udim(ustring udimpattern,
+                                        float s, float t) = 0;
+
+    /// A more efficient variety of `resolve_udim()` for cases where you
+    /// have the `TextureHandle*` that corresponds to the "virtual" UDIM
+    /// file and optionally have a `Perthread*` for the calling thread.
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual TextureHandle* resolve_udim(TextureHandle* udimfile,
+                                        Perthread* thread_info,
+                                        float s, float t) = 0;
+
+    /// Produce a full inventory of the set of concrete files comprising the
+    /// UDIM set specified by UTF-8 encoded `udimpattern`.  The apparent
+    /// number of texture atlas tiles in the u and v directions will be
+    /// written to `nutiles` and `nvtiles`, respectively. The vector
+    /// `filenames` will be sized to `ntiles * nvtiles` and filled with the
+    /// the names of the concrete files comprising the atlas, with an empty
+    /// ustring corresponding to any unpopulated tiles (the UDIM set is
+    /// allowed to be sparse). The filename list is indexed as `utile + vtile
+    /// * nvtiles`.
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual void inventory_udim(ustring udimpattern,
+                                std::vector<ustring>& filenames,
+                                int& nutiles, int& nvtiles) = 0;
+
+    /// A more efficient variety of `inventory_udim()` for cases where you
+    /// have the `TextureHandle*` that corresponds to the "virtual" UDIM
+    /// file and optionally have a `Perthread*` for the calling thread.
+    ///
+    /// This method was added in OpenImageIO 2.3.
+    virtual void inventory_udim(TextureHandle* udimfile,
+                                Perthread* thread_info,
+                                std::vector<ustring>& filenames,
+                                int& nutiles, int& nvtiles) = 0;
+    /// @}
+
     /// @{
     /// @name Controlling the cache
     ///
 
-    /// Invalidate any cached information about the named file, including
-    /// loaded texture tiles from that texture, and close any open file
-    /// handle associated with the file. This calls
-    /// `ImageCache::invalidate(filename,force)` on the underlying
-    /// ImageCache.
+    /// Invalidate any cached information about the named file (UTF-8
+    /// encoded), including loaded texture tiles from that texture, and close
+    /// any open file handle associated with the file. This calls
+    /// `ImageCache::invalidate(filename,force)` on the underlying ImageCache.
     virtual void invalidate (ustring filename, bool force = true) = 0;
 
     /// Invalidate all cached data for all textures.  This calls
     /// `ImageCache::invalidate_all(force)` on the underlying ImageCache.
     virtual void invalidate_all (bool force=false) = 0;
 
-    /// Close any open file handles associated with a named file, but do not
-    /// invalidate any image spec information or pixels associated with the
-    /// files.  A client might do this in order to release OS file handle
-    /// resources, or to make it safe for other processes to modify textures
-    /// on disk.  This calls `ImageCache::close(force)` on the underlying
-    /// ImageCache.
+    /// Close any open file handles associated with a UTF-8 encoded filename,
+    /// but do not invalidate any image spec information or pixels associated
+    /// with the files.  A client might do this in order to release OS file
+    /// handle resources, or to make it safe for other processes to modify
+    /// textures on disk.  This calls `ImageCache::close(force)` on the
+    /// underlying ImageCache.
     virtual void close (ustring filename) = 0;
 
     /// `close()` all files known to the cache.
@@ -1673,12 +1798,15 @@ public:
 
     /// @{
     /// @name Errors and statistics
-    
-    /// If any of the API routines returned false indicating an error,
-    /// this routine will return the error string (and clear any error
-    /// flags).  If no error has occurred since the last time geterror()
-    /// was called, it will return an empty string.
-    virtual std::string geterror () const = 0;
+
+    /// Is there a pending error message waiting to be retrieved?
+    virtual bool has_error() const = 0;
+
+    /// Return the text of all pending error messages issued against this
+    /// TextureSystem, and clear the pending error message unless `clear` is
+    /// false. If no error message is pending, it will return an empty
+    /// string.
+    virtual std::string geterror(bool clear = true) const = 0;
 
     /// Returns a big string containing useful statistics about the
     /// TextureSystem operations, suitable for saving to a file or
